/* Final Project - Dragon Animation - Brian Stewart */

#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>
#include <cfloat>
#include <iostream>
#include <string>
#include <time.h>
#include <vector>

#include "model.h"
#include "key_set.h"
#include "quat.h"
#include "select_sphere.h"

#define EDIT 0
#define ANIMATE 1
#define KEYFRAME_MODE 2
#define BUFSIZE 10

using namespace std;

// GLOBALS *******************************************
bool normals_on = false;

double scale = 0.75; //size of model
double xrot = 20.0;
double yrot = -20.0;
double zrot = 0.0;

double htrans = 0;
double vtrans = 0;

int w, h;
int timer = 0;
int draw_mode = EDIT;

double lightz = 0;
bool wireframe = false;
bool dd = false; //draw dragon
bool smooth = false;
GLfloat position1[] = {-10.0, -10.0, 0.0, 0.0};

axis cur_axis = none;

struct timespec t;

model * dragon;
key_set * frames;
select_sphere * point_spheres;

//end of global declarations ****************************

void setOrthographicProjection();
void resetPerspectiveProjection();

void renderBitmapString(
		float x, 
		float y, 
		float z, 
		void *font, 
		char *string) {  
  char *c;
  glRasterPos3f(x, y,z);
  for (c=string; *c != '\0'; c++) {
    glutBitmapCharacter(font, *c);
  }
}

void drawScene (int mode)
{
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
		gluLookAt(0,0,10,0,0,0,0,1,0);
	
		GLfloat position0[] = {10.0, 20.0+lightz, 10.0, 0.0};
  		 glLightfv( GL_LIGHT0, GL_POSITION, position0 );

		glLightfv( GL_LIGHT1, GL_POSITION, position1 );

		
		//draw the dragon
		glTranslatef(htrans,0.0,0.0);
		glTranslatef(0.0,-1.0,0.0);
		glScalef(scale,scale,scale);
		glRotated(xrot,1,0,0);
		glRotated(yrot,0,1,0);
/*
		glColor3f(0.0,1.0,0.0);
		glBegin(GL_QUADS);
			glVertex3f(30.0,-3.3,30.0);
			glVertex3f(30.0,-3.3,-30.0);
			glVertex3f(-30.0,-3.3,-30.0);
			glVertex3f(-30.0,-3.3,30.0);
		glEnd();*/

		if(draw_mode != ANIMATE){
		//draw some axis
		glColor3f(0.0,0.0,1.0); //zAxis blue
		glBegin(GL_LINES);
			glVertex3f(0.0,0.0,-10.0);
			glVertex3f(0.0,0.0,10.0);
		glEnd();
		glColor3f(0.0,1.0,0.0); //yAxis green
		glBegin(GL_LINES);
			glVertex3f(0.0,10.0,0.0);
			glVertex3f(0.0,-10.0,0.0);
		glEnd();
		glColor3f(1.0,0.0,0.0); //xAxis red
		glBegin(GL_LINES);
			glVertex3f(10.0,0.0,0.0);
			glVertex3f(-10.0,0.0,0.0);
		glEnd();
		}
		glPushMatrix();

		if(mode == GL_RENDER){
			if(draw_mode == ANIMATE){
				frames->animate(timer,false);
				timer += 10;
			}else if(draw_mode == KEYFRAME_MODE){
				frames->draw_keyframe();
			}
			if (dd) dragon->draw_model(normals_on, smooth);
		}
	
		if(draw_mode == EDIT){
			frames->draw_skeleton(mode);
		}
		
		if(draw_mode != ANIMATE) point_spheres->draw_spheres(mode, frames, *frames); 

		glPopMatrix();

		glFlush();

		glPopMatrix();
	//deal with the text

	setOrthographicProjection();

	glPushMatrix();
	glLoadIdentity();
	//renderBitmapString(0,1.7,0,(void *)GLUT_BITMAP_HELVETICA_18,"3D Tech");
	glPopMatrix();

	resetPerspectiveProjection();

	glutSwapBuffers(); 

	clock_gettime(CLOCK_MONOTONIC, &t);
	

}

void reshape(int w1, int h1)
{

	w = w1;
	h = h1;
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	float ratio = 1.0* w / h;

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	// Set the correct perspective.
	gluPerspective(45,ratio,1,1000);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
/*	gluLookAt(0.0,0.0,5.0, 
		      0.0,0.0,-1.0,
			  0.0f,1.0f,0.0f); */
}

void init (void) 
{
	glShadeModel(GL_SMOOTH);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0f);

    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_CULL_FACE);
    //glCullFace(GL_BACK);

    glDepthFunc(GL_LEQUAL);

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);


    // lighting stuff
    GLfloat ambient[] = {0.0, 0.0, 0.0, 0.0};
    GLfloat diffuse0[] = {1.0, 0.7, 0.0, 0.7};
    GLfloat diffuse1[] = {0.0, 0.4, 1.0, 0.7};
    GLfloat specular[] = {1.0, 1.0, 1.0, 1.0};
  
    GLfloat position0[] = {10.0, 30.0, 10.0, 0.0};
    glLightfv( GL_LIGHT0, GL_POSITION, position0 );
    glLightfv( GL_LIGHT0, GL_AMBIENT, ambient );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse1 );
    glLightfv( GL_LIGHT0, GL_SPECULAR, specular );

    GLfloat position1[] = {10.0, 30.0, 10.0, 0.0};
    glLightfv( GL_LIGHT1, GL_POSITION, position1 );
    glLightfv( GL_LIGHT1, GL_AMBIENT, ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse0 );
    glLightfv( GL_LIGHT1, GL_SPECULAR, specular );

    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glEnable( GL_LIGHT1 );
    glEnable( GL_COLOR_MATERIAL );

	glEnable(GL_NORMALIZE);

	//load dragon model
	dragon = new model((const char *)"dragon.obj");
	cout << "dragon loaded" << endl;
	//load a skeleton
	frames = new key_set(dragon);
	frames->load_skeleton("skeleton");
	
	frames->pre_compute_quat(true);
	cout << "skeleton loaded" << endl;
	point_spheres = new select_sphere(dragon,frames);
	cout << "end of init" << endl;
	

}

void display(void)
{
	GLint mode = GL_RENDER;
	glGetIntegerv(GL_RENDER_MODE, &mode);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	
	glClearColor (0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	GLfloat position1[] = {0.0, 0.0, 0.0, 0.0};
    glLightfv( GL_LIGHT1, GL_POSITION, position1 );


	drawScene (mode); //GL_RENDER
	glFlush();

	if (mode == GL_RENDER)
		glutSwapBuffers();
}



void specialKey( int key, int x, int y )
{
	switch (key) {
	  case GLUT_KEY_UP:
		  xrot -= 1;
		  //xrot %= 360;
		  break;
	  case GLUT_KEY_DOWN:
		  xrot += 1;
		  //xrot %= 360;
		  break;
	  case GLUT_KEY_LEFT:
		  yrot -= 1;
		 // yrot %= 360;
		  break;
	  case GLUT_KEY_RIGHT:
		  yrot += 1;
		 // yrot %= 360;
		  break;
	}
	glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key)
	{
	case 27: /* escape key*/
		exit(0);
		break;
	case 'w': // reset
		if(!wireframe){
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			wireframe = true;
		}else{
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			wireframe = false;
		}
		break;
	case '-':
		if(cur_axis == none)
			scale -= .1;
		else if (cur_axis == X)
			frames->change_rot(cur_axis, -1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == Y)
			frames->change_rot(cur_axis, -1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == Z)
			frames->change_rot(cur_axis, -1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == L)
			frames->change_len(-.1);
		else if (cur_axis == I)
			frames->change_dist(-.1);
		else if (cur_axis == D)
			frames->change_dur(-1);
		break;
	case 'n':
		normals_on = !normals_on;
		break;
	case '+':
		if(cur_axis == none)
			scale += .1;
		else if (cur_axis == X)
			frames->change_rot(cur_axis, 1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == Y)
			frames->change_rot(cur_axis, 1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == Z)
			frames->change_rot(cur_axis, 1, draw_mode==KEYFRAME_MODE);
		else if (cur_axis == L)
			frames->change_len(.1);
		else if (cur_axis == I)
			frames->change_dist(.1);
		else if (cur_axis == D)
			frames->change_dur(1);
		break;
	case 'a':
		lightz -= 1;
		break;
	case 's':
		cur_axis = none;
		break;
	case 'c':
		frames->change_bone();
		break;
	case 'd':
		dd = !dd;
		break;
	case 'x':
		cur_axis = X;
		break;
	case 'y':
		cur_axis = Y;
		break;
	case 'z':
		cur_axis = Z;
		break;
	case 'l':
		cur_axis = L;
		break;
	case 'f':
		frames->next_frame();
		frames->set_cur_to_keyframe();
		break;
	case 't':
		timer++;
		break;
	case 'i':
	//	cur_axis = I;
		break;
	case 'h':
		htrans += .1;
		break;
	case 'k':
		htrans -= .1;
		break;
	case 'e':
		cur_axis = D;
		break;
	}
	glutPostRedisplay();
}

void setOrthographicProjection() {

	// switch to projection mode
	glMatrixMode(GL_PROJECTION);

	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();

	// reset matrix
	glLoadIdentity();

	// set a 2D orthographic projection
	gluOrtho2D(0, w, h, 0);
	
	// switch back to modelview mode
	glMatrixMode(GL_MODELVIEW);
}

void resetPerspectiveProjection() {

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void processMenuEvents(int option) {

	switch (option) {
		case 1 : 
		 	frames->dump_skeleton("skeleton", 0);
			break;
		case 2 : 
		 	frames->append_to_anim("anim");
			break;
		case 3 :
			frames->load_animation("anim");
			frames->pre_compute_quat(false);
			break;
		case 4:
			draw_mode = ANIMATE;
			break;
		case 5:
			draw_mode = EDIT;
			frames->draw_skeleton(0);
			timer = 0;
			break;
		case 6:
			draw_mode = KEYFRAME_MODE;
			frames->set_cur_to_keyframe();
			timer = 0;
			break;
		case 7:
			frames->compute_bone_points();
			break;
		case 8:
			frames->save_animation("anim");
			break;
	}
}
void createGLUTMenus() {

	int menu;

	menu = glutCreateMenu(processMenuEvents);
	glutAddMenuEntry("Save Skeleton",1);
	glutAddMenuEntry("Append to Animation File", 2);
	glutAddMenuEntry("Load Animation File", 3);
	glutAddMenuEntry("Animate Mode", 4);
	glutAddMenuEntry("Edit Mode", 5);
	glutAddMenuEntry("Keyframe Mode", 6);
	glutAddMenuEntry("Recalculate Bone Points", 7);
	glutAddMenuEntry("Save Animation", 8);
	
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void processHits (GLint hits,GLuint buffer[]){
	int i, j, names;
   GLuint *ptr;

   printf ("hits = %d\n", hits);
   ptr = (GLuint *) buffer;
   for (i = 0; i < hits; i++) {	/*  for each hit  */
      names = (int) *ptr;
      printf (" number of names for hit = %d\n", names); ptr++;
      printf("  z1 is %g;", (float) *ptr/0x7fffffff); ptr++;
      printf(" z2 is %g\n", (float) *ptr/0x7fffffff); ptr++;
      for (j = 0; j < names; j++, ptr++) {	/*  for each name */
		  printf ("  hit %d's name is: %d\n", i, *ptr);
		  frames->change_point(dragon, *ptr);
	  }
   }

	
}

void selectObjects(int button, int state, int x, int y)
{
   GLuint selectBuf[BUFSIZE];
   GLint hits;
   GLint viewport[4];

   if (button != GLUT_LEFT_BUTTON || state != GLUT_DOWN)
      return;

   glGetIntegerv (GL_VIEWPORT, viewport);

   glSelectBuffer (BUFSIZE, selectBuf);
   (void) glRenderMode (GL_SELECT);

   glInitNames();
   glPushName(0);

   glMatrixMode (GL_PROJECTION);
   glPushMatrix ();
   glLoadIdentity ();
/*  create 2x2 pixel picking region near cursor location	*/
   gluPickMatrix ((GLdouble) x, (GLdouble) (viewport[3] - y), 
                  2.0, 2.0, viewport);
   gluPerspective(45,1.0* w / h,1,1000);
   
   display();

   glMatrixMode (GL_PROJECTION);
   glPopMatrix ();
   glFlush ();

   glPopName();

   hits = glRenderMode (GL_RENDER);
   processHits (hits, selectBuf);
}

/*  Main Loop  */
int main(int argc, char** argv)
{
   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutInitWindowSize (750, 600);
   glutInitWindowPosition (100, 100);
   glutCreateWindow (argv[0]);
 	glutMouseFunc(selectObjects);
   glutReshapeFunc(reshape);
	glutIdleFunc(display);
   init();
   glutDisplayFunc(display);
   glutKeyboardFunc(keyboard);
   glutSpecialFunc(specialKey);

	createGLUTMenus();

   glutMainLoop();
   return 0; 
}
