#include<vector>
#include<iostream>
#include<fstream>
#include<string>
#include <GL/glut.h>
#include <math.h>
#include"key_set.h"
#include"quat.h"
#include"model.h"

#define BONE_DIST .75    //distance from bone that points are influenced by bone's movement
#define FRAME_LEN 1000 //default keyframe duration

using namespace std;

key_set::key_set(model * m):cur(NULL), cframe(0), skeleton(NULL), frame_time(0), skin(m)
{
	for(unsigned i = 0;i<m->points.size();i++){
		point * p = new point;
		p->x = m->points[i].x;
		p->y = m->points[i].y;
		p->z = m->points[i].z;
		orig_skin.push_back(*p);
	}
}

//add bone: if parent is NULL, a new skeleton is added to the
//keyframe vector
bone * key_set::add_bone(bone * parent, point begin, double length, double anglex,
		double angley, double anglez, string name){
	bone * t;
	
	//if there is no parent then create a new root
	if(parent == NULL){
		if(!(parent = new bone)) return NULL;
		skeleton = parent;
		parent->parent = NULL;
		cur = parent;
	}else{ 
		if(!(t = new bone)) return NULL;
		t->parent = parent;
		parent->child.push_back(t);
		parent = t;
	}

	parent->begin.x = begin.x;
	parent->begin.y = begin.y;
	parent->begin.z = begin.z;
	parent->anglex = anglex;
	parent->angley = angley;
	parent->anglez = anglez;
	parent->curx = anglex;
	parent->cury = angley;
	parent->curz = anglez;
	parent->length = length;
	parent->anim = false;
	parent->done = true;
	parent->dist = BONE_DIST;
	parent->points.clear();
	parent->orig_points.clear();

	parent->name = name;
	
	return parent;
}

void key_set::draw_skeleton(int mode){
	//diable lighting for skeleton
	glDisable(GL_LIGHTING);
	//set id's to starting num, skin->points.size() so it doesnt interfer with spheres
	id_count = skin->points.size();
	traverse_skeleton(skeleton->begin, skeleton, mode);
	glEnable(GL_LIGHTING);
	//redraw the dragon in the right place
	move_points_rec(skeleton);
}

void key_set::traverse_skeleton(point begin, bone * root, int mode){ //recursive code to draw skeleton
	//calculate end position
	vec end;
	end.x = 0.0;
	end.y = root->length;
	end.z = 0.0;

	end = root->q * end; //do rotation based on quaternions

	end.x += begin.x;
	end.y += begin.y;
	end.z += begin.z;

	if(root == cur) glColor3f(1.0,0.0,0.0);
	else glColor3f(1.0,1.0,1.0);

	/* stuff for picking */
	if(mode == GL_SELECT)
			glLoadName(id_count);
	root->id = id_count;

	glBegin(GL_LINES);
		glVertex3f(begin.x,begin.y,begin.z);
		glVertex3f(end.x,end.y,end.z);
	glEnd();
	

	glPushMatrix();
		glTranslatef(end.x,end.y,end.z);
		glutSolidSphere(0.1, 5, 5);
	glPopMatrix(); 


	point pend; //convert vec struct back to point struct
	pend.x = end.x;
	pend.y = end.y;
	pend.z = end.z;

	id_count++;

	for(unsigned int i=0;i<root->child.size();i++){
		traverse_skeleton(pend, root->child[i], mode);
	}
}

//compute quaternions before running animation to save time
void key_set::pre_compute_quat(bool first){
	//computer for base skeleton

	quat q(0,0,0,1);
	pre_compute_quat_rec(q, skeleton->begin, skeleton, false, 0, first);

	//compute for each keyframe
	for(int i=0;i<keyframes.size();i++){
		pre_compute_quat_rec(quat(0,0,0,1), skeleton->begin, skeleton, true, i, false);
	}
}

//recurisve function for pre_compute_quat
//if keyframe, will computer quaternions for keyframe # k
void key_set::pre_compute_quat_rec(quat r, point begin, bone * root, bool keyframe, int k, bool first){
	if(k > root->keyframes.size() - 1) {
		cerr << "error: pre_compute_quat_rec: k not in bounds" << endl;
		exit(1);
	}

	// assign starting position of bone
	if(keyframe){
		root->keyframes[k]->begin.x = begin.x;
		root->keyframes[k]->begin.y = begin.y;
		root->keyframes[k]->begin.z = begin.z;
	}else{
		root->begin.x = begin.x;
		root->begin.y = begin.y;
		root->begin.z = begin.z;		
	}

	if(first){
		root->start.x = begin.x;
		root->start.y = begin.y;
		root->start.z = begin.z;
	}

	quat q;

	if(keyframe){
		q.from_euler(root->keyframes[k]->anglex,root->keyframes[k]->angley,
					root->keyframes[k]->anglez);
		q = r * q;
		root->keyframes[k]->q = q;
	}else{
		q.from_euler(root->curx,root->cury,root->curz);
		q = r * q;
		root->q = q;
		if(first) root->o = q;
	}	
	
	vec end;
	end.x = 0.0;
	end.y = root->length;
	end.z = 0.0;

	end = q * end; //do rotation based on quaternions

	end.x += begin.x;
	end.y += begin.y;
	end.z += begin.z;

	point pend; //convert vec struct back to point struct
	pend.x = end.x;
	pend.y = end.y;
	pend.z = end.z;

	for(unsigned int i=0;i<root->child.size();i++){
		pre_compute_quat_rec(q, pend, root->child[i], keyframe, k, first);
	}
}


void key_set::change_bone(){

	if(cur->child.size()>0){
		cur = cur->child[0];
		cur->done = true;
	}else{
		cur = get_next_child(cur);
	}

	//debug:

	if(cur == NULL) cur = skeleton;
	cout << "bone: " << cur->name << endl;
	if (cur->parent == NULL) cout << "	parent = NULL" << endl;
	else cout << "	parent = " << cur->parent->name << endl;
	cout << "	children: " << endl;
	for(unsigned int i=0;i<cur->child.size();i++){
		cout << "	child " << i << " " << cur->child[i]->name << endl;
	}

}

bone * key_set::get_next_child(bone * cur){
	if(cur->done == true){
		bone * tmp = cur->parent;
		for(unsigned int i=0;i<tmp->child.size();i++){
			if(tmp->child[i] == cur){
				if(i == tmp->child.size()-1){ //if last child
					cur = get_next_child(cur->parent);	
				}else{
					cur = tmp->child[i+1];
					cur->done = true;
				}
				break;
			}
		}		
	}
	return cur;
}

void key_set::change_rot(axis a, double deg, bool k){
	
	cout << "bool k = " << k << endl;

	if(cur == skeleton) return;

	if(a == X){
		if(k) cur->keyframes[cframe]->anglex += deg;
		else cur->curx += deg;
	}
	if(a == Y){
		if(k) cur->keyframes[cframe]->angley += deg;
		else cur->cury += deg;
	}
	if(a == Z){
		if(k) cur->keyframes[cframe]->anglez += deg;
		else cur->curz += deg;
	}

	

	if(k){
		for(int i=0;i<keyframes.size();i++)
			pre_compute_quat_rec(quat(0,0,0,1), skeleton->begin, skeleton, true, i, false);
		draw_keyframe();
	}else{
		move_points_rec(skeleton);
		pre_compute_quat_rec(cur->parent->q, cur->begin, cur, false, 0, false);
	}
}

void key_set::change_len(double len){
	cur->length += len;
}

bone * key_set::find_by_name_rec(bone * root, string name, int id){
	
	if(root == NULL) return NULL;

	if(root->name.compare(name) == 0 || id == root->id){ //if names are the same
		return root;
	}else{
		int count = root->child.size();
		bone * tmp;
		for(int i = 0;i<count;i++){
			tmp = find_by_name_rec(root->child[i], name, id);
			if(tmp!=NULL) return tmp;
		}
		return NULL;
	}
}

void key_set::dump_skeleton(const char * title, int f){
	ofstream out;
	out.open(title);
	dump_tree(skeleton, out);
}

void key_set::dump_tree(bone * root,ofstream & out){
	//name x, y, z, anglex, angley, anglez, length, parent
	out << root->name << " ";
	out << root->begin.x << " ";
	out << root->begin.y << " ";
	out << root->begin.z << " ";
	out << root->anglex << " ";
	out << root->angley << " ";
	out << root->anglez << " ";
	out << root->length << " ";
	if(root->parent == NULL)
		out << root->parent << endl;
	else
		out << root->parent->name  << endl;

	//now print out skin info:
	//first total size of point array
	out << root->orig_points.size() << " ";
	//then each point index
	for(unsigned i=0;i<root->orig_points.size();i++){
		out << root->orig_points[i] << " ";
	}
	out << endl;

	for(unsigned i =0;i<root->child.size();i++){
		dump_tree(root->child[i], out);
	}

}

void key_set::append_to_anim(const char * title){
	ofstream file;
	file.open (title, ios::out | ios::app);
	dump_anim(skeleton, file);
	file << endl;
	file.close();
	frame_time += FRAME_LEN;
}

void key_set::dump_anim(bone * root, ofstream & out){
	//name anglex, angley, anglez, length, time
	out << root->name << " ";
	out << root->curx << " ";
	out << root->cury << " ";
	out << root->curz << " ";
	out << root->length << " ";
	out << frame_time << endl;

	for(unsigned i =0;i<root->child.size();i++){
		dump_anim(root->child[i], out);
	}
}

void key_set::save_animation(const char * title){
	ofstream file;
	file.open (title);

	for(unsigned i=0;i<keyframes.size();i++){
		dump_anim_key(skeleton, file, i);
		file << endl;
	}	

	file.close();
}

void key_set::dump_anim_key(bone * root, ofstream & out, int k){
	//name anglex, angley, anglez, length, time
	out << root->name << " ";
	out << root->keyframes[k]->anglex << " ";
	out << root->keyframes[k]->angley << " ";
	out << root->keyframes[k]->anglez << " ";
	out << root->keyframes[k]->length << " ";
	out << keyframes[k] << endl;

	for(unsigned i =0;i<root->child.size();i++){
		dump_anim_key(root->child[i], out, k);
	}
}

//loads a skeleton
void key_set::load_skeleton(const char * title){

	ifstream file (title);
	if (file.is_open()){
   	while ( file.good() )
  		{
			string parent;
			bone tmp;
      	file >> tmp.name;
			file >> tmp.begin.x;
			file >> tmp.begin.y;
			file >> tmp.begin.z;
			file >> tmp.anglex;
			file >> tmp.angley;
			file >> tmp.anglez;
			file >> tmp.length;
			if(!file.good()) break;
			file >> parent;
			tmp.parent = find_by_name_rec(skeleton, parent, -1);

			bone * b = add_bone(tmp.parent, tmp.begin, tmp.length, tmp.anglex, tmp.angley, 
				tmp.anglez, tmp.name);

			//now load in bone points
			int p;
			file >> p;
			for(int i=0;i<p;i++){
				int t;
				file >> t;
				b->points.push_back(&skin->points[t]);
				b->orig_points.push_back(t);
			}
    	}
    	file.close();
	}
	else cout << "unable to open file" << endl;
	skeleton->done = false;
}

//loads an animation file
//assumes all keyframes are in order i.e. keyframes with higher times
//come after keyframes with lower times
void key_set::load_animation(const char * title){

	cout << "loading animation file..." << endl;
	ifstream file (title);
	uint32_t counter = -1;
	if (file.is_open()){
   	while ( file.good() )
  		{
			string parent, name;
			double anglex, angley, anglez, len;
			uint32_t time;
			bone * b;

      	file >> name;
			file >> anglex;
			file >> angley;
			file >> anglez;
			file >> len;
			file >> time;

			if(!file.good()) {
				break;
			}
	
			//update keyframes array
			if(keyframes.size() == 0 || time > keyframes.back())
					keyframes.push_back(time);

			b = find_by_name_rec(skeleton, name, -1);
			
			if(b == NULL){
				cout << "animation file does not correspond to skeleton!" << endl;
				exit(1);
			}
			
			keyframe * k = new keyframe;
			k->anglex = anglex;
			k->angley = angley;
			k->anglez = anglez;
			k->length = len;
			k->time = time;
			b->keyframes.push_back(k);
			
    	}
    	file.close();
	}
	else cout << "unable to open file" << endl;

	for(int i=0;i<keyframes.size();i++){
		cout << "keyframe #" << i << " is " << keyframes[i] << endl;
	}
	
	if(keyframes.size() > 0)
		frame_time = keyframes.back()+FRAME_LEN;

}

void key_set::compute_bone_points(){
	compute_bone_points_rec(quat(0,0,0,1), skeleton->begin, skeleton, skin);
}


void key_set::compute_bone_points_rec(quat r, point begin, bone * root, model * m){
	double dist;
	//calculate end position

	vec end;
	end.x = 0.0;
	end.y = root->length;
	end.z = 0.0;

	quat q;
	q.from_euler(root->anglex,root->angley,root->anglez);
	q = r * q;
	end = q * end; //do rotation based on quaternions

	end.x += begin.x;
	end.y += begin.y;
	end.z += begin.z;

	//clear points vector incase there are already some points stored
	root->points.clear();
	root->orig_points.clear();

	//now calcuate points which are close to the bone
	for(int i=0;i<m->points.size();i++){
		
		if((end.x-begin.x) * (m->points[i].x-begin.x) +
			(end.y-begin.y) * (m->points[i].y-begin.y) +
			(end.z-begin.z) * (m->points[i].z-begin.z) >= 0)
			{
			if((end.x-begin.x) * (m->points[i].x-end.x) +
				(end.y-begin.y) * (m->points[i].y-end.y) +
				(end.z-begin.z) * (m->points[i].z-end.z) <= 0)
				{
				vec v1, v2;
				v1.x = (m->points[i].x - begin.x);
				v1.y = (m->points[i].y - begin.y);
				v1.z = (m->points[i].z - begin.z);
		
				v2.x = (m->points[i].x - end.x);
				v2.y = (m->points[i].y - end.y);
				v2.z = (m->points[i].z - end.z);

				//cross product
				vec num;
				num.x = v1.y * v2.z - v1.z * v2.y;
				num.y = v1.z * v2.x - v1.x * v2.z;
				num.z = v1.x * v2.y - v1.y * v2.x;
	
				vec dem;
				dem.x = end.x - begin.x;
				dem.y = end.y - begin.y;
				dem.z = end.z - begin.z;

				double num_len = sqrt(num.x*num.x+num.y*num.y+num.z*num.z);
				double dem_len = sqrt(dem.x*dem.x+dem.y*dem.y+dem.z*dem.z);
		
				if (num_len/dem_len < root->dist){
					root->points.push_back(&m->points[i]);
					root->orig_points.push_back(i);
				}
			}else{
				dist = sqrt((m->points[i].x-end.x)*(m->points[i].x-end.x)+
								(m->points[i].y-end.y)*(m->points[i].y-end.y)+
								(m->points[i].z-end.z)*(m->points[i].z-end.z));
				if(dist < root->dist){
					root->points.push_back(&m->points[i]);
					root->orig_points.push_back(i);
				}
			}
		}else{
			dist = sqrt((m->points[i].x-begin.x)*(m->points[i].x-begin.x)+
							(m->points[i].y-begin.y)*(m->points[i].y-begin.y)+
							(m->points[i].z-begin.z)*(m->points[i].z-begin.z));
			if(dist < root->dist){
				root->points.push_back(&m->points[i]);
				root->orig_points.push_back(i);
			}
		}
	}

	point pend; //convert vec struct back to point struct
	pend.x = end.x;
	pend.y = end.y;
	pend.z = end.z;

	for(unsigned int i=0;i<root->child.size();i++){
		compute_bone_points_rec(q, pend, root->child[i], m);
	}
}

void key_set::change_dist(double a){
	cur->dist += a;
	cout << "cur->dist = " << cur->dist << endl;
	compute_bone_points_rec(quat(0,0,0,1), skeleton->begin, skeleton, skin);
}

void key_set::next_frame(){
	cframe++;
	if (cframe >= keyframes.size()) cframe = 0;
	cout << "keyframe switched to " << cframe << endl;
}

void key_set::draw_keyframe(){
	if ( keyframes.size() == 0)
		cout << "error: no keyframes uploaded" << endl;
	else
		animate(keyframes[cframe], true);

	
}

void key_set::animate(int t, bool draw){
	//compute which keyframes to interpolate between
	glDisable(GL_LIGHTING);
	int prev, next;
	double u;

	if(keyframes.size() < 2) {
		cerr << "need at least two keyframes to animate" << endl;
		return;
	}

	//cout << "t = " << t << "keyframes.back() = " << keyframes.back() << endl;
	if (t > keyframes.back()) t = (t%keyframes.back()); //repeat animation if it gets to end

	for(unsigned i=1;i<keyframes.size();i++){
		if(t <= keyframes[i]) {
			prev = i-1;
			next = i;
			break;
		}
	}
	 
	u = t-keyframes[prev];
	u = u/(keyframes[next]-keyframes[prev]);

	//create a identity quaternion to start with
	quat q(0,0,0,1);
	animate_rec(q, skeleton->begin, skeleton, prev, next, u, draw);

	glEnable(GL_LIGHTING);
}

void key_set::set_cur_to_keyframe(){
	set_cur_to_keyframe_rec(skeleton);
	pre_compute_quat_rec(quat(0,0,0,1), skeleton->begin, skeleton, false, 0, false);
}


void key_set::set_cur_to_keyframe_rec(bone * root){

	root->curx = root->keyframes[cframe]->anglex;
	root->cury = root->keyframes[cframe]->angley;
	root->curz = root->keyframes[cframe]->anglez;

	for(unsigned i=0;i<root->child.size();i++){
		set_cur_to_keyframe_rec(root->child[i]);
	}
}

void key_set::animate_rec(    //recursive code to animate skeleton //calculate end position
	quat last, 
	point begin, 
	bone * root, 
	int prev, 
	int next, 
	double u,
	bool draw)
{ 
	vec end;
	end.x = 0.0;
	end.y = root->length;
	end.z = 0.0;

	quat r;
	//do linear interp
	//r = (root->keyframes[prev]->q * (1-u)) + (root->keyframes[next]->q * u);
 	r.slerp(root->keyframes[prev]->q, root->keyframes[next]->q, u);
	end = r * end; //do rotation based on final quaternion

	end.x += begin.x;
	end.y += begin.y;
	end.z += begin.z;

	if(draw){
	if(root == cur) glColor3f(1.0,0.0,0.0);
	glBegin(GL_LINES);
		glVertex3f(begin.x,begin.y,begin.z);
		glVertex3f(end.x,end.y,end.z);
	glEnd();

	glPushMatrix();
		glTranslatef(end.x,end.y,end.z);
		glutSolidSphere(0.1, 5, 5);
	glPopMatrix(); 
	glColor3f(1.0,1.0,1.0);
	}

	point pend; //convert vec struct back to point struct
	pend.x = end.x;
	pend.y = end.y;
	pend.z = end.z;

	move_points_anim(root, r, begin);

	for(unsigned int i=0;i<root->child.size();i++){
		animate_rec(r, pend, root->child[i], prev, next, u, draw);
	}
}

void key_set::move_points_rec(bone * root){

	for(int i=0;i<root->points.size();i++){

		vec v;
		v.x = orig_skin[root->orig_points[i]].x - root->start.x;
		v.y = orig_skin[root->orig_points[i]].y - root->start.y;
		v.z = orig_skin[root->orig_points[i]].z - root->start.z;

		//quat a = root->o.get_conjugate().get_partial_rot(.5);
		v = root->o.get_conjugate() * v;

		//a = root->q.get_partial_rot(.5);
		quat w;
		w.slerp(root->o, root->q, 1);

		v = w * v; //do rotation based on quaternions
		
		root->points[i]->x = v.x + root->begin.x;
		root->points[i]->y = v.y + root->begin.y;
		root->points[i]->z = v.z + root->begin.z;
	}

	for(int i = 0;i<root->child.size();i++){
		move_points_rec(root->child[i]);
	} 

}


void key_set::move_points_anim(bone * root, quat & q, point & begin){

	for(int i=0;i<root->points.size();i++){

		vec v;
		v.x = orig_skin[root->orig_points[i]].x - root->start.x;
		v.y = orig_skin[root->orig_points[i]].y - root->start.y;
		v.z = orig_skin[root->orig_points[i]].z - root->start.z;

		v = root->o.get_conjugate() * v;

		v = q * v; //do rotation based on quaternions
		
		root->points[i]->x = v.x + begin.x;
		root->points[i]->y = v.y + begin.y;
		root->points[i]->z = v.z + begin.z;
	}
}


void key_set::change_point(model * m, int index){
	
	//first check to see if its a sphere hit or a bone hit
	if(index < m->points.size()){
		cout << "changing bone point" << endl;
		point * p = &m->points[index];

		//check if point is already assigned to the bone
		for(int i=0;i<cur->points.size();i++){

			if(p == cur->points[i]){
				cur->points.erase(cur->points.begin()+i);
				cur->orig_points.erase(cur->orig_points.begin()+i);
				return;
			}
		}

		cur->points.push_back(p);
		cur->orig_points.push_back(index);
	
		move_points_rec(skeleton);

		//debug
		cout << "cur.points = " << endl;
		for(int i=0;i<cur->points.size();i++){
			cout << cur->orig_points[i] << endl;
		}
	}else{
		string s = "";
		bone * b = find_by_name_rec(skeleton,s,index);
		if(b == NULL) cout << "picking error" << endl;
		else cur = b;
	}
}

void key_set::change_dur(double a){
	if(cframe >= keyframes.size()-1) return;
	cout << "duration for keyframe " << cframe << " until next keyframe changed to "
		<< skeleton->keyframes[cframe+1]->time + a << endl;
	for(unsigned i=cframe+1;i<keyframes.size();i++){
		keyframes[i] += a;
		change_dur_rec(skeleton, i, a);
	}
}

void key_set::change_dur_rec(bone* root, int k, double a){

	root->keyframes[k]->time += a;

	if(root->keyframes[k]->time < 0) root->keyframes[k]->time = 0;

	for(unsigned i=0;i<root->child.size();i++){
		change_dur_rec(root->child[i], k, a);
	}
}

