#ifndef KEY_SET_H
#define KEY_SET_H

#include<vector>
#include<iostream>
#include<fstream>
#include<string>
#include<stdint.h>
#include"model.h"
#include"quat.h"

typedef struct keyframe{
	double length;
									
	point begin;   //starting point of bone (should be preconputed)

	/*angles of rotation*/
	double anglex;
	double angley;
	double anglez;
	quat q;
	
	uint32_t time; //time that this keyframe happens
} keyframe;

typedef struct bone
 {
 	std::string name;				/* name of bone*/
	int id;                  //used for picking
	point begin;				//where the bone starts
	point start;              //location of bone relative to mesh
	double length;			   //length of bone
									/*angles of rotation*/
	double anglex;
	double angley;
	double anglez;

	quat o; //original quat

	//current angles 
	double curx;
	double cury;
	double curz;

	bool anim;

	double dist;         //the distance that the bone has influence over mesh points

	quat q;              //current quaternion for making smooth animations
	bool done;				//used for traversing the skeleton
 
	std::vector<struct point *>  points; /* points that the bone is connected to*/
	std::vector<int>  orig_points; /* the indexes of the original points values */
 	std::vector<struct bone *> child;	/* Pointers to children */
 	struct bone *parent;			/* Parent bone */
	std::vector<struct keyframe *> keyframes; //the keyframes of the bone
 } bone;

enum axis {X, Y, Z, none, L, I, D};

class key_set
{
public:
	key_set(model *);
	~key_set();
	void dump_skeleton(const char * title, int f);
	void draw_skeleton(int);
	void change_bone();
	void change_rot(axis a, double deg, bool k);
	void change_len(double deg);
	bone * add_bone(bone * parent, point begin, double length, double anglex,
		double angley, double anglez, std::string name);
	void load_skeleton(const char * title);
	void load_animation(const char * title);
	void append_to_anim(const char * title);
	void compute_bone_points();	//compute the bone points by distance
	void draw_keyframe();  
	void next_frame();	
	void animate(int t, bool draw);
	void pre_compute_quat(bool);
	void change_point(model * m, int);
	void save_animation(const char * title);
	void change_dist(double a);
	void set_cur_to_keyframe();
	void change_dur(double a); //change the duration of the keyframe

	//fields
	bone * cur;
	bone * skeleton;
	std::vector<uint32_t> keyframes; //array of time values for each keyframe

private:
	void traverse_skeleton(point begin, bone * root, int);
	bone * find_by_name_rec(bone * root, std::string name, int);
	void dump_tree(bone * root, std::ofstream&);
	void dump_anim(bone * root, std::ofstream & out);
	bone * get_next_child(bone * cur);
	void animate_rec(quat last, point begin, bone * root, int prev, int next, double u, bool draw);
	void compute_bone_points_rec(quat r, point begin, bone * root, model * m);
	void pre_compute_quat_rec(quat r, point begin, bone * root, bool keyframe, int k, bool first);
	void move_points_rec(bone * root);
	void move_points_anim(bone * root, quat & q, point & begin);
	void dump_anim_key(bone * root, std::ofstream & out, int k);
	void set_cur_to_keyframe_rec(bone * root);
	void change_dur_rec(bone* root, int k, double a);

	//fields
	model * skin;
	int id_count;
	std::vector<point> orig_skin; //vector array of the original points
											//(used when morphing the skin)
	int cframe; //tells which keyframe to draw if using draw_keyframe();
	uint32_t frame_time; //keeps track of the time for each keyframe that is saved

	friend class select_sphere;
	
};

#endif
